import random
from typing import Any
import pygame
from PIL import Image
import gymnasium as gym
import numpy as np
from gymnasium import spaces

import gym2048.environment as e


pygame.init()


class Game2048(gym.Env):
    def __init__(self, binary=False, seed=None, display=False, save_animation=None):
        """
        Initialize a new 2048 game environment.

        @param binary: Whether to one hot encode the board. Useful for neural networks.
        @param seed: The random seed to use. Defaults to a computer generated seed.
        @param display: Whether to display the changes in a PyGame window.
        @param save_animation: Whether to save the displayed animation to a gif.
        """
        super(Game2048, self).__init__()

        self.binary = binary
        self.display = display
        self.save_animation = save_animation if self.display else False
        self.animation = []
        self.rng = np.random.default_rng(seed)

        # Create the action space.
        self.action_space = spaces.Discrete(4, seed=self.rng)

        # Create the observation space.
        self.board = e.initialize(self.rng)

        if self.binary:
            self.observation_space = spaces.Box(0, 1, (4, 4, 12), seed=self.rng)
        else:
            self.observation_space = spaces.Box(0, 2048, (4, 4), seed=self.rng)

        if self.display:
            self._initialize_render()

    def step(self, action):
        """
        Take a step with the given action.

        * Reward of 1 if the game was won.
        * Reward of 0 if the game was lost.
        * Reward of -0.1 if the agent made an invalid move.

        @param action: The action to take: (0: up, 1: down, 2: left, 3: right).
        @return: (observation, reward, terminated, truncated, None)
        """
        reward = 0
        done = False

        if e.game_won(self.board) or e.game_lost(self.board):
            raise Exception("Environment has terminated. Call reset!")

        if action not in e.possible_moves(self.board)[0]:
            reward = 0
            done = False  # The game ends if the board
        else:
            self.board = e.move(self.board, action, self.rng)
            if e.game_won(self.board):
                reward = 1
            else:
                reward = 0

        next_obs = self.board if not self.binary else e.to_onehot(self.board)
        done = e.game_won(self.board) or e.game_lost(self.board) or done

        if self.display:
            self.render()

        # Return the observation, reward, terminated, truncated, extra info
        return next_obs, reward, done, False, {"board": np.copy(self.board)}

    def reset(self, *, seed: int | None = None, options: dict[str, Any] | None = None):
        """
        Reset the environment.

        @param seed: The new seed to use. Old generator will continue otherwise.
        @param options: N/A
        @return: (observation, None)
        """
        if seed is not None:
            self.rng = np.random.default_rng(seed)
        self.board = e.initialize(self.rng)
        obs = self.board if not self.binary else e.to_onehot(self.board)
        return obs, {"board": np.copy(self.board)}

    def _draw_grid(self):
        for row in range(4):
            for col in range(4):
                value = self.board[row][col]
                color = self.tile_colors.get(value, self.tile_colors[2048])
                pygame.draw.rect(
                    self.screen,
                    color,
                    (
                        col * self.cell_size,
                        row * self.cell_size,
                        self.cell_size,
                        self.cell_size,
                    ),
                )
                if value != 0:
                    font = pygame.font.Font(None, 36)
                    text = font.render(str(value), True, (0, 0, 0))
                    text_rect = text.get_rect(
                        center=(
                            col * self.cell_size + self.cell_size / 2,
                            row * self.cell_size + self.cell_size / 2,
                        )
                    )
                    self.screen.blit(text, text_rect)

    def _initialize_render(self):
        self.window_size = (400, 400)  # Width, Height
        self.cell_size = 100

        self.tile_colors = {
            0: (255, 255, 255),
            2: (238, 228, 218),
            4: (237, 224, 200),
            8: (242, 177, 121),
            16: (245, 149, 99),
            32: (246, 124, 95),
            64: (246, 94, 59),
            128: (237, 207, 114),
            256: (237, 204, 97),
            512: (237, 200, 80),
            1024: (237, 197, 63),
            2048: (237, 194, 46),
        }

        self.screen = pygame.display.set_mode(self.window_size)
        pygame.display.set_caption("2048 Game")

    def render(self):
        """
        Print the current board state to the terminal.
        """
        self.screen.fill((255, 255, 255))
        self._draw_grid()
        pygame.display.flip()

        if self.save_animation:
            self.animation.append(
                pygame.surfarray.array3d(pygame.display.get_surface())
            )

    def save_render(self, filename):
        """
        Save the animation if possible. Requires the environment to have been
        initialized with the save_animation flag marked True.

        @param filename: The filename to save the generated gif to.
        """
        if len(self.animation) > 2:
            image_list = [Image.fromarray(arr) for arr in self.animation]
            image_list[0].save(
                filename,
                save_all=True,
                append_images=image_list[1:],
                duration=100,  # Duration between frames in milliseconds
                loop=0,
            )
            self.animation = []


if __name__ == "__main__":
    game = Game2048(display=True, save_animation=False)
    obs, _ = game.reset()
    done = False

    while not done:
        _, _, term, trunc, _ = game.step(random.randint(0, game.action_space.n))
        done = term or trunc

    game.save_render("output.gif")
    pygame.quit()



import sys

import numpy as np


def add_tile(board, rng):
    """
    Add a tile to the given board and return it.

    :param board: The board represented as a numpy array.
    :param rng: The random number generator.
    :return: The updated board.
    """
    board = np.copy(board)
    empty_squares = np.argwhere((board == 0))
    if len(empty_squares) == 0:
        return board
    idx = empty_squares[rng.integers(0, len(empty_squares))]
    num = rng.choice(np.array([2, 4]), p=np.array([0.9, 0.1]))
    board[idx[0], idx[1]] = num
    return board


def initialize(rng):
    """
    Create a new board.

    :param rng: The random number generator.
    :return: A new board array.
    """
    board = np.zeros((4, 4))
    board = add_tile(board, rng)
    board = add_tile(board, rng)
    return board


def up(board, rng):
    original_board = np.copy(board)
    board = np.copy(board)
    s = board.shape[0]
    for j in range(s):
        section = board.T[j]
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        for i in range(s - 1):
            if section[i] == section[i + 1] and section[i] != 0:
                section[i] = section[i] + section[i + 1]
                section[i + 1] = 0
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        board[:, j] = section
    if np.array_equal(original_board, board):
        return original_board, False
    board = add_tile(board, rng)
    return board, True


def down(board, rng):
    original_board = np.copy(board)
    board = np.copy(board)
    board = np.flipud(board)
    s = board.shape[0]
    for j in range(s):
        section = board.T[j]
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        for i in range(s - 1):
            if section[i] == section[i + 1] and section[i] != 0:
                section[i] = section[i] + section[i + 1]
                section[i + 1] = 0
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        board[:, j] = section
    board = np.flipud(board)
    if np.array_equal(original_board, board):
        return original_board, False
    board = add_tile(board, rng)
    return board, True


def left(board, rng):
    original_board = np.copy(board)
    board = np.copy(board)
    s = board.shape[0]
    for j in range(s):
        section = board[j]
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        for i in range(s - 1):
            if section[i] == section[i + 1] and section[i] != 0:
                section[i] = section[i] + section[i + 1]
                section[i + 1] = 0
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        board[j, :] = section
    if np.array_equal(original_board, board):
        return original_board, False
    board = add_tile(board, rng)
    return board, True


def right(board, rng):
    original_board = np.copy(board)
    board = np.copy(board)
    board = np.fliplr(board)
    s = board.shape[0]
    for j in range(s):
        section = board[j]
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        for i in range(s - 1):
            if section[i] == section[i + 1] and section[i] != 0:
                section[i] = section[i] + section[i + 1]
                section[i + 1] = 0
        for i in range(s - 1):
            if section[i] == 0:
                section[i:-1] = section[i + 1 :]
                section[-1] = 0
        board[j, :] = section
    board = np.fliplr(board)
    if np.array_equal(original_board, board):
        return original_board, False
    board = add_tile(board, rng)
    return board, True


def game_won(board):
    return bool(np.isin(2048, board))


def game_lost(board):
    rng = np.random.default_rng(0)
    return not game_won(board) and not (
        up(board, rng)[1] == True
        or down(board, rng)[1] == True
        or left(board, rng)[1] == True
        or right(board, rng)[1] == True
    )


def move(board, action, rng):
    """
    Perform an action on the given board.

    * board: The board to make the move on.
    * action: The action to take 0: up, 1: down, 2: left, 3: right
    * rng: The random generator

    * return: The new board.
    """
    if action == 0:
        board, _ = up(board, rng)
    if action == 1:
        board, _ = down(board, rng)
    if action == 2:
        board, _ = left(board, rng)
    if action == 3:
        board, _ = right(board, rng)
    return board


def status(board):
    if game_won(board):
        return 1
    elif game_lost(board):
        return -1
    else:
        return 0


def score(board):
    return np.max(board)


def possible_moves(board):
    rng = np.random.default_rng()
    moves = []
    boards = []
    up_board, up_bool = up(board, rng)
    down_board, down_bool = down(board, rng)
    left_board, left_bool = left(board, rng)
    right_board, right_bool = right(board, rng)
    if up_bool:
        moves.append(0)
        boards.append(up_board)
    if down_bool:
        moves.append(1)
        boards.append(down_board)
    if left_bool:
        moves.append(2)
        boards.append(left_board)
    if right_bool:
        moves.append(3)
        boards.append(right_board)
    return np.array(moves), boards


def to_onehot(board):
    log_board = np.log2(board, where=(board > 0))
    unique_values = int(np.log2(2048)) + 1
    one_hot = np.zeros((*board.shape, unique_values))
    try:
        for x, y in np.argwhere(board > 0):
            one_hot[x, y, int(np.log2(board[x, y]))] = 1
    except IndexError:
        print(board)
        sys.exit()
    return one_hot


def from_onehot(one_hot):
    tmp = np.argmax(one_hot, axis=-1)
    tmp = 2**tmp
    tmp[tmp == 1] = 0
    print(tmp)
    return tmp
